/*----------------------------------------------------------------------
  File    : JNIFIM.c
  Contents: Frequent Item Set Mining for Java
  Author  : Christian Borgelt
  History : 2014.10.08 file created from file JNICoCo.c
            2014.10.09 auxiliary functions programmed
            2014.10.15 frequent item set mining functions completed
            2014.10.16 pattern spectrum functions completed
            2014.10.17 first version completed and debugged
            2015.02.28 item appearances param. added to rule functions
            2015.03.01 bug in report string evaluation fixed (jchar)
            2015.03.03 signal handling added (interrupt/ctrl-c)
            2015.08.07 ensured signal handler removal on error
            2015.08.28 simple parsing of /proc/cpuinfo added
            2016.11.05 adapted to modified apriori interface
            2016.11.11 adapted to modified eclat interface
            2016.11.15 adapted to modified accretion interface
            2016.11.21 adapted to modified fpgrowth interface
            2018.03.21 Jaccard Item set Mining added (JIM)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#include <pthread.h>
#endif
#ifndef TATREEFN
#define TATREEFN
#endif
#ifndef TA_SURR
#define TA_SURR
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#ifndef PSP_ESTIM
#define PSP_ESTIM
#endif
#include "sigint.h"
#include "report.h"
#include "apriori.h"
#include "eclat.h"
#include "fpgrowth.h"
#include "sam.h"
#include "relim.h"
#include "jim.h"
#include "carpenter.h"
#include "ista.h"
#include "accretion.h"
#include "fpgpsp.h"
#include "JNIFIM.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE     1024        /* block size for arrays */

#define CMSETS      (ISR_CLOSED|ISR_MAXIMAL)
#define FCMSETS     (CMSETS|ISR_SETS)
#define ANYSETS     (FCMSETS|ISR_GENERAS)
#define SETSvRULES  (ANYSETS|ISR_RULES)

/* --- error handling --- */
#define X_MEM       0           /* OutOfMemoryError */
#define X_NULL      1           /* NullPointerException */
#define X_ARG       2           /* IllegalArgumentException */
#define X_CLASS     3           /* ClassNotFoundException */
#define X_ABORT     4           /* InterruptedException */
#define X_NONE      5           /* dummy */

#define ERR_RTN(e)      error(e, X_NONE,  NULL)
#define ERR_MEM(e)      error(e, X_MEM,   "not enough memory")
#define ERR_NULL(e)     error(e, X_NULL,  "tracts")
#define ERR_ARG(e,s)    error(e, X_ARG,   s)
#define ERR_CLS(e,s)    error(e, X_CLASS, s)
#define ERR_ABORT(e)    error(e, X_ABORT, "user abort")

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- item set/rule report data --- */
  JNIEnv       *env;            /* JNI environment */
  jobjectArray res;             /* result for Java function */
  jsize        size;            /* size of item set/ass. rule array */
  jsize        cnt;             /* number of item sets/ass. rules */
  jint         *heads;          /* head items of association rules */
  jintArray    *sets;           /* frequent item sets/rule bodies */
  jsize        len;             /* number of value indicators */
  void         **info;          /* add. item set/rule information */
  char         *rep;            /* indicators of values to report */
  char         *sois;           /* sizes of info: sizeof(int/double) */
  int          err;             /* error flag */
} REPDATA;                      /* (item set/rule report data) */

typedef struct {                /* --- Java environment data --- */
  JNIEnv       *env;            /* Java Native Interface environment */
  jint         done[1];         /* number of completed data sets */
  jintArray    ctrl;            /* Java control variables */
} JDATA;                        /* (Java environment data) */

/*----------------------------------------------------------------------
  Parameter Functions
----------------------------------------------------------------------*/

static int get_app (jint app)
{                               /* --- get item appearance code */
  switch (app) {                /* evaluate the item appearance */
    case fim_JNIFIM_IGNORE:     return APP_NONE;
    case fim_JNIFIM_BODY:       return APP_BODY;
    case fim_JNIFIM_HEAD:       return APP_HEAD;
    case fim_JNIFIM_BOTH:       return APP_BOTH;
  }                             /* return the item appearance code */
  return -1;                    /* return an error code */
}  /* get_app() */

/*--------------------------------------------------------------------*/

static int get_target (jint target, int sel)
{                               /* --- get target code */
  switch (target) {             /* evaluate the target type */
    case fim_JNIFIM_FREQUENT:   target = ISR_SETS;    break;
    case fim_JNIFIM_CLOSED:     target = ISR_CLOSED;  break;
    case fim_JNIFIM_MAXIMAL:    target = ISR_MAXIMAL; break;
    case fim_JNIFIM_GENERAS:    target = ISR_GENERAS; break;
    case fim_JNIFIM_RULES:      target = ISR_RULES;   break;
    default:                   return -1;
  }                             /* return the target code */
  if ((target != 0) && ((target & sel) == 0))
    return -1;                  /* check for a valid target type */
  return target;                /* return an error code */
}  /* get_target() */

/*--------------------------------------------------------------------*/

static int get_stat (jint stat)
{                               /* --- get statistic code */
  switch (stat) {               /* evaluate the statistic code */
    case fim_JNIFIM_NONE:       return RE_NONE;
    case fim_JNIFIM_CHI2PVAL:   return RE_CHI2PVAL;
    case fim_JNIFIM_YATESPVAL:  return RE_YATESPVAL;
    case fim_JNIFIM_INFOPVAL:   return RE_INFOPVAL;
    case fim_JNIFIM_FETPROB:    return RE_FETPROB;
    case fim_JNIFIM_FETCHI2:    return RE_FETCHI2;
    case fim_JNIFIM_FETINFO:    return RE_FETINFO;
    case fim_JNIFIM_FETSUPP:    return RE_FETSUPP;
  }                             /* return the statistic code */
  return -1;                    /* return an error code */
}  /* get_stat() */

/*--------------------------------------------------------------------*/

static int get_eval (jint eval)
{                               /* --- get evaluation measure code */
  switch (eval) {               /* evaluate evaluation measure code */
    case fim_JNIFIM_NONE:       return RE_NONE;
    case fim_JNIFIM_LDRATIO:    return RE_FNCNT;
  }                             /* return the evaluation measure code */
  return -1;                    /* return an error code */
}  /* get_eval() */

/*--------------------------------------------------------------------*/

static int get_evalx (jint eval)
{                               /* --- get evaluation measure code */
  switch (eval) {               /* evaluate evaluation measure code */
    case fim_JNIFIM_NONE:       return RE_NONE;
    case fim_JNIFIM_SUPP:       return RE_SUPP;
    case fim_JNIFIM_CONF:       return RE_CONF;
    case fim_JNIFIM_CONFDIFF:   return RE_CONFDIFF;
    case fim_JNIFIM_LIFT:       return RE_LIFT;
    case fim_JNIFIM_LIFTDIFF:   return RE_LIFTDIFF;
    case fim_JNIFIM_LIFTQUOT:   return RE_LIFTQUOT;
    case fim_JNIFIM_CVCT:       return RE_CVCT;
    case fim_JNIFIM_CVCTDIFF:   return RE_CVCTDIFF;
    case fim_JNIFIM_CVCTQUOT:   return RE_CVCTQUOT;
    case fim_JNIFIM_CPROB:      return RE_CPROB;
    case fim_JNIFIM_IMPORT:     return RE_IMPORT;
    case fim_JNIFIM_CERT:       return RE_CERT;
    case fim_JNIFIM_CHI2:       return RE_CHI2;
    case fim_JNIFIM_CHI2PVAL:   return RE_CHI2PVAL;
    case fim_JNIFIM_YATES:      return RE_YATES;
    case fim_JNIFIM_YATESPVAL:  return RE_YATESPVAL;
    case fim_JNIFIM_INFO:       return RE_INFO;
    case fim_JNIFIM_INFOPVAL:   return RE_INFOPVAL;
    case fim_JNIFIM_FETPROB:    return RE_FETPROB;
    case fim_JNIFIM_FETCHI2:    return RE_FETCHI2;
    case fim_JNIFIM_FETINFO:    return RE_FETINFO;
    case fim_JNIFIM_FETSUPP:    return RE_FETSUPP;
    case fim_JNIFIM_LDRATIO:    return RE_FNCNT;
  }                             /* return the evaluation measure code */
  return -1;                    /* return an error code */
}  /* get_evalx() */

/*--------------------------------------------------------------------*/

static int get_agg (jint agg)
{                               /* --- get aggregation mode code */
  switch (agg) {                /* evaluate aggregation mode code */
    case fim_JNIFIM_NONE:       return IST_NONE;
    case fim_JNIFIM_MIN:        return IST_MIN;
    case fim_JNIFIM_MAX:        return IST_MAX;
    case fim_JNIFIM_AVG:        return IST_AVG;
  }                             /* return the aggregation mode code */
  return -1;                    /* return an error code */
}  /* get_agg() */

/*--------------------------------------------------------------------*/

static int get_covsim (jint covsim)
{                               /* --- get cover similarity code */
  switch (covsim) {             /* evaluate cover similarity code */
    case fim_JNIFIM_JIM_NONE:            return JIM_NONE;
    case fim_JNIFIM_JIM_RUSSEL_RAO:      return JIM_RUSSEL_RAO;
    case fim_JNIFIM_JIM_KULCYNSKI:       return JIM_KULCZYNSKI;
    case fim_JNIFIM_JIM_JACCARD:         return JIM_JACCARD;
    case fim_JNIFIM_JIM_DICE:            return JIM_DICE;
    case fim_JNIFIM_JIM_SOKAL_SNEATH_1:  return JIM_SOKAL_SNEATH_1;
    case fim_JNIFIM_JIM_HAMMING:         return JIM_HAMMING;
    case fim_JNIFIM_JIM_FAITH:           return JIM_FAITH;
    case fim_JNIFIM_JIM_ROGERS_TANIMOTO: return JIM_ROGERS_TANIMOTO;
    case fim_JNIFIM_JIM_SOKAL_SNEATH_2:  return JIM_SOKAL_SNEATH_2;
    case fim_JNIFIM_JIM_SOKAL_SNEATH_3:  return JIM_SOKAL_SNEATH_3;
    case fim_JNIFIM_JIM_BARONI_BUSER:    return JIM_BARONI_BUSER;
    case fim_JNIFIM_JIM_GENERIC:         return JIM_GENERIC;
  }                             /* return the cover similarity code */
  return -1;                    /* return an error code */
}  /* get_covsim() */

/*--------------------------------------------------------------------*/

static int get_surr (jint surr)
{                               /* --- get surrogate method code */
  switch (surr) {               /* evaluate the surrogate method code */
    case fim_JNIFIM_IDENT:      return 0;
    case fim_JNIFIM_RANDOM:     return 1;
    case fim_JNIFIM_SWAP:       return 2;
    case fim_JNIFIM_SHUFFLE:    return 3;
  }                             /* return the surrogate method code */
  return -1;                    /* return an error code */
}  /* get_surr() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void* error (JNIEnv *env, int type, const char *msg)
{                               /* --- throw an exception */
  jclass cls;                   /* class of the exception */
  char   *name;                 /* name of the exception */

  assert(env);                  /* check the function arguments */
  sig_remove();                 /* remove the signal handler */
  switch (type) {               /* evaluate the exception type */
    case X_MEM:   name = "java/lang/OutOfMemoryError";         break;
    case X_NULL:  name = "java/lang/NullPointerException";     break;
    case X_ARG:   name = "java/lang/IllegalArgumentException"; break;
    case X_CLASS: name = "java/lang/ClassNotFoundException";   break;
    case X_ABORT: name = "java/lang/InterruptedException";     break;
    default:      name = NULL;
  }                             /* get the name and then the class */
  if (type != X_NONE) {         /* if to raise an exception */
    cls = (*env)->FindClass(env, name);
    (*env)->ThrowNew(env, cls, (msg) ? msg : "");
  }                             /* create and throw exception */
  return NULL;                  /* throw the exception and return */
}  /* error() */

/*--------------------------------------------------------------------*/

static ITEMBASE* ib_appJObj (JNIEnv *env, ITEMBASE *ibase,
                             jobjectArray appear)
{                               /* --- get item appearances */
  int     e = 0;                /* error indicator */
  jobject o0, o1;               /* to access the Java arrays */
  jsize   k, n;                 /* loop variable, array lengths */
  jint    *items;               /* array of items */
  jint    *apps;                /* array of item apearances */
  ITEM    i;                    /* item identifier */
  int     a;                    /* item appearance indicator */

  assert(env && ibase);         /* check the function arguments */
  if (!appear) return ibase;    /* if no item apps. given, abort */
  if ((*env)->GetArrayLength(env, appear) < 2)
    return ERR_ARG(env, "appear must have at least two elements");
  o0 = (*env)->GetObjectArrayElement(env, appear, 0);
  if (!o0) {                    /* get the item array */
    (*env)->DeleteLocalRef(env, o0);
    return ERR_ARG(env, "missing item array in 'appear");
  }                             /* check whether item array exists */
  o1 = (*env)->GetObjectArrayElement(env, appear, 1);
  if (!o1) {                    /* get the item appearances array */
    (*env)->DeleteLocalRef(env, o0);
    (*env)->DeleteLocalRef(env, o1);
    return ERR_ARG(env, "missing item appearances in 'appear'");
  }                             /* check whether apps array exists */
  n = (*env)->GetArrayLength(env, o0);
  k = (*env)->GetArrayLength(env, o1);
  if (n != k) {                 /* get the array lengths */
    (*env)->DeleteLocalRef(env, o0);
    (*env)->DeleteLocalRef(env, o1);
    return ERR_ARG(env, "arrays in 'appear' must have same length");
  }                             /* check for equal array lengths */
  items = (*env)->GetIntArrayElements(env, o0, NULL);
  apps  = (*env)->GetIntArrayElements(env, o1, NULL);
  for (k = 0; k < n; k++) {     /* traverse the arrays */
    i = (ITEM)items[k];         /* get the next item */
    if (i >= 0) {               /* if normal item (not default) */
      i = ib_add(ibase, &i);    /* add item to the item base */
      if (i < -1) continue;     /* if item already exists, skip it */
      if (i <  0) { e = -1; break; }
    }                           /* if not enough memory, abort */
    a = get_app(apps[k]);       /* get the appearance indicator */
    if (a < 0) { e = 1; break; }
    ib_setapp(ibase, i, a);     /* set the appearance indicator */
  }                             /* for the current item */
  (*env)->ReleaseIntArrayElements(env, o0, items, JNI_ABORT);
  (*env)->ReleaseIntArrayElements(env, o1, apps,  JNI_ABORT);
  (*env)->DeleteLocalRef(env, o0);
  (*env)->DeleteLocalRef(env, o1);
  if (e > 0) return ERR_ARG(env, "invalid item appearance indicator");
  if (e < 0) return ERR_MEM(env);
  return ibase;                 /* return the modified item base */
}  /* ib_appJObj() */

/*--------------------------------------------------------------------*/

static TABAG* tbg_fromJObj (JNIEnv *env, jobjectArray tracts,
                            jintArray wgts, jobjectArray appear)
{                               /* --- create a transaction bag */
  int      e = 0;               /* error flag */
  jsize    i, j, k, n;          /* loop variables */
  jobject  o;                   /* to traverse the Java arrays */
  jint     *t;                  /* items   of the transactions */
  jint     *w;                  /* weights of the transactions */
  ITEM     item;                /* to traverse the items */
  ITEMBASE *ibase;              /* underlying item base */
  TABAG    *tabag;              /* created transaction bag */

  assert(env && tracts);        /* check the function argument */
  n = (*env)->GetArrayLength(env, tracts);
  if (wgts && ((*env)->GetArrayLength(env, wgts) != n))
    return ERR_ARG(env, "tracts and wgts must have same length");
  ibase = ib_create(IB_OBJNAMES, 0, ST_INTFN, (OBJFN*)0);
  if (!ibase) return ERR_MEM(env);   /* create an item base */
  if (!ib_appJObj(env, ibase, appear)) {
    ib_delete(ibase); return NULL; } /* add item appearances */
  tabag = tbg_create(ibase);    /* create a transaction bag */
  if (!tabag) { ib_delete(ibase); return ERR_MEM(env); }
  w = (wgts) ? (*env)->GetIntArrayElements(env, wgts, NULL) : NULL;
  for (i = 0; i < n; i++) {     /* traverse the transactions */
    ib_clear(ibase);            /* clear the current transaction */
    o = (*env)->GetObjectArrayElement(env, tracts, i);
    k = (*env)->GetArrayLength(env, o);
    t = (*env)->GetIntArrayElements(env, o, NULL);
    for (j = 0; j < k; j++) {   /* traverse the items and */
      item = (ITEM)t[j];        /* add them to the item base */
      if (ib_add2ta(ibase, &item) < 0) { e = -1; break; }
    }                           /* (built trans. in item base */
    (*env)->ReleaseIntArrayElements(env, o, t, JNI_ABORT);
    (*env)->DeleteLocalRef(env, o);
    if (e) break;               /* release the Java array */
    ib_finta(ibase, (w) ? (SUPP)w[i] : 1);
    e = tbg_addib(tabag);       /* terminate the transaction and */
    if (e) break;               /* add the transaction to the bag */
  }                             /* and check for an error */
  if (w) (*env)->ReleaseIntArrayElements(env, wgts, w, JNI_ABORT);
  if (e) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  return tabag;                 /* return the created transaction bag */
}  /* tbg_fromJObj() */

/*--------------------------------------------------------------------*/

static int isr_Jborder (JNIEnv *env, ISREPORT *rep, jintArray border)
{                               /* --- set reporter filtering border */
  jsize n;                      /* loop variable, border length */
  jint  *b;                     /* elements of the border array */

  assert(env && rep);           /* check the function arguments */
  if (!border) return 0;        /* abort if no border given */
  n = (*env)->GetArrayLength(env, border);
  if (n <= 0)  return 0;        /* empty borders need no processing */
  b = (*env)->GetIntArrayElements(env, border, NULL);
  while (n > 0) { --n;          /* traverse the border elements */
    if (isr_setbdr(rep, (ITEM)n, (RSUPP)b[n]) < 0)
      break;                    /* set the support values and */
  }                             /* afterward release Java array */
  (*env)->ReleaseIntArrayElements(env, border, b, JNI_ABORT);
  return (n <= 0) ? 0 : -1;     /* return an error status */
}  /* isr_Jborder() */

/*--------------------------------------------------------------------*/

static void isr_iset2JObj (ISREPORT *rep, void *data)
{                               /* --- report an item set */
  REPDATA   *rd = data;         /* type the data pointer */
  jsize     k, n;               /* loop variables, array size */
  ITEM      i, m;               /* loop variables, number of items */
  RSUPP     supp, base;         /* item set support and base support */
  SUPP      s;                  /* support value for reporting */
  jdouble   e, x;               /* evaluation and scaling factor */
  void*     p;                  /* resized array (sets of info) */
  jint      *set;               /* item set elements */
  jintArray jset;               /* new Java item set */

  assert(rep && data);          /* check the function arguments */
  if (rd->err) return;          /* if there was an error, do nothing */
  n = rd->size;                 /* get the current array size */
  if (rd->cnt >= n) {           /* if the item set array is full */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    p = realloc(rd->sets, (size_t)n *sizeof(jintArray));
    if (!p) { rd->err = -1; return; }
    rd->sets = p;               /* enlarge the item set array */
    for (k = 0; k < rd->len; k++) {
      p = realloc(rd->info[k], (size_t)n *(size_t)rd->sois[k]);
      if (!p) { rd->err = -1; return; }
      rd->info[k] = p;          /* traverse and enlarge */
    }                           /* the information arrays */
    rd->size = n;               /* set the new array size */
  }
  m    = isr_cnt(rep);          /* create object for item set */
  jset = (*rd->env)->NewIntArray(rd->env, (jsize)m);
  if (!jset) { rd->err = -1; return; }
  set  = (*rd->env)->GetIntArrayElements(rd->env, jset, NULL);
  for (i = 0; i < m; i++)       /* map identifiers to items */
    set[i] = (jint)(ptrdiff_t)isr_itemobj(rep, isr_itemx(rep, i));
  (*rd->env)->ReleaseIntArrayElements(rd->env, jset, set, 0);
  rd->sets[rd->cnt] = jset;     /* store item set and its support */
  supp = isr_supp (rep);        /* get the item set support */
  base = isr_suppx(rep, 0);     /* and the total transaction weight */
  s = 0; e = 0;                 /* initialize the report variables */
  for (k = 0; k < rd->len; k++){/* traverse the values to store */
    switch (rd->rep[k]) {       /* evaluate the value indicator */
      case 'a': s = (SUPP)supp;                   x =   0.0; break;
      case 's': e = (jdouble)supp /(jdouble)base; x =   1.0; break;
      case 'S': e = (jdouble)supp /(jdouble)base; x = 100.0; break;
      case 'p': e = isr_eval(rep);                x =   1.0; break;
      case 'P': e = isr_eval(rep);                x = 100.0; break;
      case 'e': e = isr_eval(rep);                x =   1.0; break;
      case 'E': e = isr_eval(rep);                x = 100.0; break;
      case 'w': e = isr_wgt(rep);                 x =   1.0; break;
      case 'W': e = isr_wgt(rep);                 x = 100.0; break;
      case 'Q': s = (SUPP)base;                   x =   0.0; break;
      default : s = 0;                            x =   0.0; break;
    }                           /* get the requested value */
    if (x == 0) ((jint*)   rd->info[k])[rd->cnt] = (jint)s;
    else        ((jdouble*)rd->info[k])[rd->cnt] = x*e;
  }                             /* store the requested value */
  rd->cnt++;                    /* count the reported item set */
}  /* isr_iset2JObj() */

/*--------------------------------------------------------------------*/

static double lift (RSUPP supp, RSUPP body, RSUPP head, RSUPP base)
{                               /* --- compute lift value of a rule */
  return ((body <= 0) || (head <= 0)) ? 0
       : ((double)supp*(double)base) /((double)body*(double)head);
}  /* lift() */

/*--------------------------------------------------------------------*/

static void isr_rule2JObj (ISREPORT *rep, void *data,
                           ITEM item, RSUPP body, RSUPP head)
{                               /* --- report an association rule */
  REPDATA   *rd = data;         /* report data structure */
  jsize     k, n;               /* loop variables, array size */
  ITEM      i, m;               /* loop variables, number of items */
  ITEM      z;                  /* to traverse the items */
  RSUPP     supp, base;         /* item set support and base support */
  SUPP      s;                  /* support value for reporting */
  double    e, x;               /* evaluation and scaling factor */
  void*     p;                  /* resized array */
  jint      *set;               /* body item set elements */
  jintArray jset;               /* new Java item set for rule body */

  assert(rep && data            /* check the function arguments */
  &&    (body > 0) && (head > 0));
  assert(isr_uses(rep, item));  /* head item must be in item set */
  if (rd->err) return;          /* if there was an error, do nothing */
  n = rd->size;                 /* get the current array size */
  if (rd->cnt >= n) {           /* if the item set array is full */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    p = realloc(rd->heads, (size_t)n *sizeof(jint));
    if (!p) { rd->err = -1; return; }
    rd->heads = p;              /* enlarge the head item array */
    p = realloc(rd->sets,  (size_t)n *sizeof(jintArray));
    if (!p) { rd->err = -1; return; }
    rd->sets = p;               /* enlarge the body set array */
    for (k = 0; k < rd->len; k++) {
      p = realloc(rd->info[k], (size_t)n *(size_t)rd->sois[k]);
      if (!p) { rd->err = -1; return; }
      rd->info[k] = p;          /* traverse and enlarge */
    }                           /* the information arrays */
    rd->size = n;               /* set the new array size */
  }                             /* store the head item of the rule */
  rd->heads[rd->cnt] = (jint)(ptrdiff_t)isr_itemobj(rep, item);
  m    = isr_cnt(rep);          /* create object for body item set */
  jset = (*rd->env)->NewIntArray(rd->env, (jsize)(m-1));
  if (!jset) { rd->err = -1; return; }
  set  = (*rd->env)->GetIntArrayElements(rd->env, jset, NULL);
  for (k = 0, i = 0; i < m; i++) {   /* map identifiers to items */
    z = isr_itemx(rep, i);      /* get the next item and skip it */
    if (z == item) continue;    /* if it is the head of the rule */
    set[k++] = (jint)(ptrdiff_t)isr_itemobj(rep, z);
  }                             /* store the item in the rule body */
  (*rd->env)->ReleaseIntArrayElements(rd->env, jset, set, 0);
  rd->sets[rd->cnt] = jset;     /* store item set and its support */
  supp = isr_supp (rep);        /* get the item set support */
  base = isr_suppx(rep, 0);     /* and the total transaction weight */
  s = 0; e = 0;                 /* initialize the report variables */
  for (i = 0; i < rd->len; i++){/* traverse the values to store */
    switch (rd->rep[i]) {       /* evaluate the value indicator */
      case 'a': s = (SUPP)supp;                   x =   0.0; break;
      case 'b': s = (SUPP)body;                   x =   0.0; break;
      case 'h': s = (SUPP)head;                   x =   0.0; break;
      case 's': e = (jdouble)supp /(jdouble)base; x =   1.0; break;
      case 'S': e = (jdouble)supp /(jdouble)base; x = 100.0; break;
      case 'x': e = (jdouble)body /(jdouble)base; x =   1.0; break;
      case 'X': e = (jdouble)body /(jdouble)base; x = 100.0; break;
      case 'y': e = (jdouble)head /(jdouble)base; x =   1.0; break;
      case 'Y': e = (jdouble)head /(jdouble)base; x = 100.0; break;
      case 'c': e = (jdouble)supp /(jdouble)body; x =   1.0; break;
      case 'C': e = (jdouble)supp /(jdouble)body; x = 100.0; break;
      case 'l': e = lift(supp, body, head, base); x =   1.0; break;
      case 'L': e = lift(supp, body, head, base); x = 100.0; break;
      case 'e': e = isr_eval(rep);                x =   1.0; break;
      case 'E': e = isr_eval(rep);                x = 100.0; break;
      case 'Q': s = (SUPP)base;                   x =   0.0; break;
      default : s = 0;                            x =   0.0; break;
    }                           /* get the requested value */
    if (x == 0) ((jint*)   rd->info[i])[rd->cnt] = (jint)s;
    else        ((jdouble*)rd->info[i])[rd->cnt] = x*e;
  }                             /* store the requested value */
  rd->cnt++;                    /* count the reported assoc. rule */
}  /* isr_rule2JObj() */

/*--------------------------------------------------------------------*/

static jobjectArray psp_toJObj (JNIEnv *env, PATSPEC *psp,
                                double scale, int format)
{                               /* --- report pattern spectrum */
  ITEM         z;               /* loop variable for sizes */
  SUPP         s;               /* loop variable for support values */
  size_t       n, k;            /* buffer for number of occurrences */
  jsize        i;               /* array index */
  jclass       cls;             /* class of integer arrays */
  jmethodID    cid;             /* id of constructor method */
  jintArray    jsizes, jsupps;  /* sizes and support values */
  jint         *sizes, *supps;  /* ditto, array elements */
  jdoubleArray jfrqs;           /* occcurence frequencies */
  jdouble      *frqs;           /* ditto, array elements */
  jobjectArray jpsp;            /* pattern spectrum as Java object */
  jobject      jpse;            /* pattern spectrum element */

  assert(env && psp);           /* check the function arguments */
  n = psp_sigcnt(psp);          /* get the number of signatures */
  if ((format == '=')           /* if list of objects (triplets) */
  ||  (format == '-')) {        /* as objects PatSpecElem */
    cls  = (*env)->FindClass(env, "java/lang/Object");
    jpsp = (*env)->NewObjectArray(env, (jsize)n, cls, NULL);
    if (!jpsp) return ERR_MEM(env); /* create a PatSpecElem array */
    cls  = (*env)->FindClass (env, "fim/PatSpecElem");
    if (!cls)  return ERR_CLS(env, "fim/PatSpecElem");
    cid  = (*env)->GetMethodID(env, cls, "<init>", "(IID)V");
    if (!cid)  return ERR_CLS(env, "fim/PatSpecElem");
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        jpse = (*env)->NewObject(env, cls, cid, (jint)z, (jint)s,
                                 (jdouble)((double)k *scale));
        if (!jpse) return NULL; /* create a pattern spectrum element */
        (*env)->SetObjectArrayElement(env, jpsp, i, jpse); i++;
      }                         /* store the pattern spectrum element */
    } }                         /* and advance the array index */
  else {                        /* if three arrays (size,supp,freq) */
    jsizes = (*env)->NewIntArray   (env, (jsize)n);
    jsupps = (*env)->NewIntArray   (env, (jsize)n);
    jfrqs  = (*env)->NewDoubleArray(env, (jsize)n);
    cls    = (*env)->FindClass(env, "java/lang/Object");
    jpsp   = (*env)->NewObjectArray(env, 3, cls, NULL);
    if (!jsizes || !jsupps || !jfrqs || !jpsp)
      return ERR_MEM(env);      /* create a Java pattern spectrum */
    (*env)->SetObjectArrayElement(env, jpsp, 0, jsizes);
    (*env)->SetObjectArrayElement(env, jpsp, 1, jsupps);
    (*env)->SetObjectArrayElement(env, jpsp, 2, jfrqs);
    sizes = (*env)->GetIntArrayElements   (env, jsizes, NULL);
    supps = (*env)->GetIntArrayElements   (env, jsupps, NULL);
    frqs  = (*env)->GetDoubleArrayElements(env, jfrqs,  NULL);
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        sizes[i]   = (jint)z; supps[i] = (jint)s;
        frqs [i++] = (jdouble)((double)k *scale);
      }                         /* store the signature elements */
    }                           /* in the corresponding arrays */
    (*env)->ReleaseIntArrayElements   (env, jsizes, sizes, 0);
    (*env)->ReleaseIntArrayElements   (env, jsupps, supps, 0);
    (*env)->ReleaseDoubleArrayElements(env, jfrqs,  frqs,  0);
  }                             /* release the arrays */
  assert((size_t)i == n);       /* check signature counter */
  return jpsp;                  /* return created pattern spectrum */
}  /* psp_toJObj() */

/*--------------------------------------------------------------------*/

static int repinit (JNIEnv *env, REPDATA *data, ISREPORT *isrep,
                    jstring report, int target)
{                               /* --- initialize reporting */
  int         i;                /* loop variable */
  const jchar *s;               /* report indicator string */
  char        *cdbl;            /* characters for double arrays */

  assert(data && isrep && report); /* check the function arguments */
  data->err  = 0;               /* initialize the error indicator */
  data->sets = NULL;            /* clear the fields that are checked */
  data->info = NULL;            /* by the function repclean() */
  s = (*env)->GetStringChars(env, report, NULL);
  if ((s[0] == '#')             /* if to get a pattern spectrum */
  ||  (s[0] == '|')             /* '#','|': three columns */
  ||  (s[0] == '=')             /* '=','-': array of PatSpecElem */
  ||  (s[0] == '-')) {
    (*env)->ReleaseStringChars(env, report, s);
    return isr_addpsp(isrep, NULL);
  }                             /* add pattern spectrum to reporter */
  data->env   = env;            /* note the JNI environment and */
  data->res   = NULL;           /* initialize the report structure */
  data->size  = data->cnt = 0;  /* there are no item sets/rules yet */
  data->heads = NULL;           /* clear head item and body set array */
  data->sets  = NULL;           /* and get the number of info chars. */
  data->len   = (int)(*env)->GetStringLength(env, report);
  data->info  = (void**)malloc((size_t)data->len    *sizeof(void*)
                              +(size_t)data->len *2 *sizeof(char));
  if (!data->info) {            /* create the reporting array */
    (*env)->ReleaseStringChars(env, report, s); return -1; }
  data->rep   = (char*)(data->info +data->len);
  data->sois  =         data->rep  +data->len;
  cdbl = (target & ISR_RULES) ? "sSxXyYcClLeEwW" : "sSpPeEwW";
  for (i = 0; i < data->len; i++) {
    data->rep [i] = (char)s[i]; /* copy the information indicators */
    data->sois[i] = (char)((strchr(cdbl, (char)s[i]) != NULL)
                           ? sizeof(jdouble) : sizeof(jint));
    data->info[i] = NULL;       /* store the data type size */
  }                             /* and clear all arrays */
  (*env)->ReleaseStringChars(env, report, s);
  if (target & ISR_RULES) isr_setrule(isrep, isr_rule2JObj, data);
  else                    isr_setrepo(isrep, isr_iset2JObj, data);
  return 0;                     /* return 'ok' */
}  /* repinit() */

/*--------------------------------------------------------------------*/

static int repterm (JNIEnv *env, REPDATA *data, ISREPORT *isrep,
                    jstring report, int target)
{                               /* --- terminate reporting */
  jsize        i, k, n;         /* loop variables, number of elements */
  jclass       cls;             /* class of object/integer arrays */
  jobjectArray jsets;           /* found item sets as Java array */
  jintArray    jia;             /* integer array for support etc. */
  jdoubleArray jda;             /* double  array for evaluation etc. */
  const jchar  *s;              /* report indicator string */
  jchar        c;               /* report indicator character */
  char         *cdbl;           /* characters for double arrays */

  assert(data && isrep && report); /* check the function arguments */
  s = (*env)->GetStringChars(env, report, NULL);
  c = s[0];                     /* get the report indicator */
  (*env)->ReleaseStringChars(env, report, s);
  if ((c == '#')                /* if to get a pattern spectrum */
  ||  (c == '|')                /* '#','|': three columns */
  ||  (c == '=')                /* '=','-': array of PatSpecElem */
  ||  (c == '-')) {
    data->res = psp_toJObj(env, isr_getpsp(isrep), 1.0, c);
    return data->err = (data->res) ? 0 : 1;
  }                             /* make Java pattern spectrum */
  cls = (*env)->FindClass(env, "java/lang/Object");
  n   = data->len;              /* get the number of output arrays */
  k   = n +((target & ISR_RULES) ? 2 : 1);
  data->res = (*env)->NewObjectArray(env, k, cls, NULL);
  if (!data->res) return -1;    /* create an object array */
  k = 0;                        /* init. the output array index */
  if (target & ISR_RULES) {     /* if to report association rules */
    jia = (*env)->NewIntArray(env, data->cnt);
    if (!jia) return -1;        /* create and copy head item array */
    (*env)->SetIntArrayRegion(env, jia, 0, data->cnt, data->heads);
    (*env)->SetObjectArrayElement(env, data->res, k, jia);
    k += 1;                     /* set the head item array and */
  }                             /* advance the array index */
  cls   = (*env)->FindClass(env, "[I");
  jsets = (*env)->NewObjectArray(env, data->cnt, cls, NULL);
  if (!jsets) return -1;        /* create the body item set array */
  for (i = 0; i < data->cnt; i++)  /* set the body item sets */
    (*env)->SetObjectArrayElement(env, jsets, i, data->sets[i]);
  (*env)->SetObjectArrayElement(env, data->res, k, jsets);
  k += 1;                       /* advance the array index */
  cdbl = (target & ISR_RULES)   /* get characters for double arrays */
       ? "sSxXyYcClLeEwW" : "sSpPeEwW";
  for (i = 0; i < n; i++) {     /* traverse the report indicators */
    if (strchr(cdbl, data->rep[i]) != NULL) {
      jda = (*env)->NewDoubleArray(env, data->cnt);
      if (!jda) return -1;      /* create and copy a double array */
      (*env)->SetDoubleArrayRegion (env, jda, 0, data->cnt,
                                    (jdouble*)data->info[i]);
      (*env)->SetObjectArrayElement(env, data->res, k, jda); }
    else {                      /* depending on the report indicator */
      jia = (*env)->NewIntArray(env, data->cnt);
      if (!jia) return -1;      /* create and copy an integer array */
      (*env)->SetIntArrayRegion    (env, jia, 0, data->cnt,
                                    (jint*)data->info[i]);
      (*env)->SetObjectArrayElement(env, data->res, k, jia);
    }                           /* store the created array */
    k += 1;                     /* advance the array index */
  }
  return data->err;             /* return the error status */
}  /* repterm() */

/*--------------------------------------------------------------------*/

static void repclean (REPDATA *data)
{                               /* --- clean up reporting data */
  jsize i;                      /* loop variable */
  if (data->sets) { free(data->sets); data->sets = NULL; }
  if (data->info) {             /* delete the information arrays */
    for (i = 0; i < data->len; i++)
      if (data->info[i]) free(data->info[i]);
    free(data->info); data->info = NULL;
  }                             /* delete the base array */
}  /* repclean() */

/*--------------------------------------------------------------------*/

static void repfn (long cnt, void *data)
{                               /* --- progress reporting function */
  JDATA  *jd = data;            /* Java environment data */
  static long buf = 0;          /* already reported progress */

  jd->done[0] = (jint)buf;      /* set the control variable */
  (*jd->env)->SetIntArrayRegion(jd->env, jd->ctrl, 0, 1, jd->done);
  if ((cnt != buf) && ((cnt % 20) == 0))
    fprintf(stderr, "%10ld\b\b\b\b\b\b\b\b\b\b", buf = cnt);
}  /* repfn() */                /* report number of data sets */

/*----------------------------------------------------------------------
  Frequent Item Set Mining Functions
----------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_fim
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jintArray border)
{                               /* --- frequent item set mining */
  int      algo  = FPG_SIMPLE;  /* algorithm variant */
  int      mode  = FPG_DEFAULT; /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  FPGROWTH *fpgrowth;           /* fpgrowth miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, ANYSETS); /* translate the target code */
  if (target < 0)    return ERR_ARG(env, "invalid target type");
  if (zmin   < 0)    return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0)    zmax = ITEM_MAX;   /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (zmin   > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax   > ITEM_MAX) zmax = ITEM_MAX;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  fpgrowth = fpg_create(target, supp, 100.0, 100.0,
                        (ITEM)zmin, (ITEM)zmax,
                        RE_NONE, FPG_NONE, 0.0, algo, mode);
  if (!fpgrowth) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = fpg_data(fpgrowth, tabag, 0, +2);
  if (r) fpg_delete(fpgrowth,1);/* prepare data for fpgrowth */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep                    /* create an item set reporter */
  ||  (fpg_report(fpgrowth, isrep) != 0)) {
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = fpg_mine(fpgrowth, ITEM_MIN, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  fpg_delete(fpgrowth, 1);      /* delete the fpgrowth miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_fim() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_xfim
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jint agg, jdouble thresh, jintArray border)
{                               /* --- frequent item set mining */
  int      algo  = FPG_SIMPLE;  /* algorithm variant */
  int      mode  = FPG_DEFAULT; /* operation mode/flags */
  long     prune = LONG_MIN;    /* min. size for evaluation filtering */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  FPGROWTH *fpgrowth;           /* fpgrowth miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, ANYSETS); /* translate the target code */
  if (target < 0)    return ERR_ARG(env, "invalid target type");
  if (zmin   < 0)    return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0)    zmax = ITEM_MAX;   /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (zmin   > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax   > ITEM_MAX) zmax = ITEM_MAX;
  eval = get_evalx(eval);       /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  if (eval   <= RE_NONE) prune = LONG_MIN;
  agg  = get_agg(agg);          /* get and check aggregation mode */
  if (agg    < 0) return ERR_ARG(env, "invalid aggregation mode");

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  fpgrowth = fpg_create(target, supp, 100.0, 100.0,
                        (ITEM)zmin, (ITEM)zmax,
                        eval, agg, thresh, algo, mode);
  if (!fpgrowth) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = fpg_data(fpgrowth, tabag, 0, +2);
  if (r) fpg_delete(fpgrowth,1);/* prepare data for fpgrowth */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  r = fpg_report(fpgrowth, isrep);
  if (r) {                      /* prepare the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (prune < ITEM_MIN) prune = ITEM_MIN;
  if (prune > ITEM_MAX) prune = ITEM_MAX;
  r = fpg_mine(fpgrowth, (ITEM)prune, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  fpg_delete(fpgrowth, 1);      /* delete the fpgrowth miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_xfim() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_arules
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jdouble supp, jdouble conf, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint jmode, jobjectArray appear)
{                               /* --- association rule mining */
  int      algo = FPG_SINGLE;   /* algorithm variant */
  int      mode = FPG_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  FPGROWTH *fpgrowth;           /* fpgrowth miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  if ((conf < 0) || (conf > 100))   /* check rule confidence */
    return ERR_ARG(env, "invalid confidence");
  if (zmin < 1) return ERR_ARG(env, "zmin must be > 0");
  if (zmax < 1) zmax = ITEM_MAX; /* check size range */
  if (zmax < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (zmin   > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax   > ITEM_MAX) zmax = ITEM_MAX;
  eval = get_evalx(eval);       /* get and check evaluation measure */
  if (eval < 0) return ERR_ARG(env, "invalid evaluation measure");
  if (jmode & fim_JNIFIM_ORIGSUPP) mode |=  APR_ORIGSUPP;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, appear);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  fpgrowth = fpg_create(FPG_RULES, supp, 100.0, conf,
                        (ITEM)zmin, (ITEM)zmax,
                        eval, FPG_NONE, thresh, algo, mode);
  if (!fpgrowth) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = fpg_data(fpgrowth, tabag, 0, +2);
  if (r) fpg_delete(fpgrowth,1);/* prepare data for fpgrowth */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  r = fpg_report(fpgrowth, isrep);
  if (r) {                      /* prepare the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  if ((repinit(env, &data, isrep, report, ISR_RULES) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }

  /* --- run association rule mining --- */
  r = fpg_mine(fpgrowth, ITEM_MIN, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, ISR_RULES);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  fpg_delete(fpgrowth, 1);      /* delete the fpgrowth miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_arules() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_apriori
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jdouble conf, jint zmin, jint zmax,
   jstring report, jint eval, jint agg, jdouble thresh, jint prune,
   jint algo, jint jmode, jintArray border, jobjectArray appear)
{                               /* --- Apriori algorithm */
  int      mode = APR_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  APRIORI  *apriori;            /* apriori miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, SETSvRULES); /* translate target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if ((conf < 0) || (conf > 100))  /* check rule confidence */
    return ERR_ARG(env, "invalid confidence");
  if (zmin  < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax  < 0) zmax = ITEM_MAX;   /* check the size range */
  if (zmax  < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_evalx(eval);       /* get and check evaluation measure */
  if (eval  < 0) return ERR_ARG(env, "invalid evaluation measure");
  if (eval  <= RE_NONE) prune = ITEM_MIN;
  if (mode & fim_JNIFIM_INVBXS) eval |= APR_INVBXS;
  agg  = get_agg(agg);          /* get and check aggregation mode */
  if (agg   < 0) return ERR_ARG(env, "invalid aggregation mode");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = APR_AUTO;  break;
    case fim_JNIFIM_APRI_BASIC:     algo = APR_BASIC; break;
    default: return ERR_ARG(env, "invalid Apriori algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~APR_PERFECT;
  if (jmode & fim_JNIFIM_NOTREE)    mode &= ~APR_TATREE;
  if (jmode & fim_JNIFIM_POSTPRUNE) mode |=  APR_POST;
  if (jmode & fim_JNIFIM_ORIGSUPP)  mode |=  APR_ORIGSUPP;

  /* --- get and prepare transactions --- */
  sig_install();                /* install the signal handler */
  if (!(target & ISR_RULES)) appear = NULL;
  tabag = tbg_fromJObj(env, tracts, wgts, appear);
  if (!tabag) return ERR_RTN(env); /* get transactions, create miner */
  apriori = apriori_create(target, supp, 100.0, conf,
                           (ITEM)zmin, (ITEM)zmax,
                           eval, agg, thresh, algo, mode);
  if (!apriori) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = apriori_data(apriori, tabag, 0, +2);
  if (r) apriori_delete(apriori, 1);   /* prepare data for apriori */
  if (r == -1) return ERR_MEM(env);    /* and check for error */
  if (r <   0) return ERR_RTN(env);    /* and for no items */

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create an item set reporter */
  if (!isrep) { apriori_delete(apriori, 1); return ERR_MEM(env); }
  r = apriori_report(apriori, isrep);  /* prepare item set reporter */
  if (r)      { apriori_delete(apriori, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    apriori_delete(apriori, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (prune < ITEM_MIN) prune = ITEM_MIN;
  if (prune > ITEM_MAX) prune = ITEM_MAX;
  r = apriori_mine(apriori, (ITEM)prune, 0.01, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  apriori_delete(apriori, 1);   /* delete the apriori miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_apriori() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_eclat
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jdouble conf, jint zmin, jint zmax,
   jstring report, jint eval, jint agg, jdouble thresh, jint prune,
   jint algo, jint jmode, jintArray border, jobjectArray appear)
{                               /* --- Eclat algorithm */
  int      mode = ECL_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  ECLAT    *eclat;              /* eclat miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, SETSvRULES); /* translate target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if ((conf < 0) || (conf > 100))  /* check rule confidence */
    return ERR_ARG(env, "invalid confidence");
  if (zmin  < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax  < 0) zmax = ITEM_MAX;   /* check the size range */
  if (zmax  < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_evalx(eval);       /* get and check evaluation measure */
  if (eval  < 0) return ERR_ARG(env, "invalid evaluation measure");
  if (eval  <= RE_NONE) prune = ITEM_MIN;
  if (mode & fim_JNIFIM_INVBXS) eval |= ECL_INVBXS;
  agg  = get_agg(agg);          /* get and check aggregation mode */
  if (agg   < 0) return ERR_ARG(env, "invalid aggregation mode");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = ECL_AUTO;   break;
    case fim_JNIFIM_ECLAT_BASIC:    algo = ECL_BASIC;  break;
    case fim_JNIFIM_ECLAT_LISTS:    algo = ECL_LISTS;  break;
    case fim_JNIFIM_ECLAT_BITS:     algo = ECL_BITS;   break;
    case fim_JNIFIM_ECLAT_TABLE:    algo = ECL_TABLE;  break;
    case fim_JNIFIM_ECLAT_SIMPLE:   algo = ECL_SIMPLE; break;
    case fim_JNIFIM_ECLAT_RANGES:   algo = ECL_RANGES; break;
    case fim_JNIFIM_ECLAT_OCCDLV:   algo = ECL_OCCDLV; break;
    case fim_JNIFIM_ECLAT_DIFFS:    algo = ECL_DIFFS;  break;
    default: return ERR_ARG(env, "invalid Eclat algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOFIM16)   mode &= ~ECL_FIM16;
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~ECL_PERFECT;
  if (jmode & fim_JNIFIM_NOHUT)     mode &= ~ECL_TAIL;
  if (jmode & fim_JNIFIM_HORZ)      mode |=  ECL_HORZ;
  if (jmode & fim_JNIFIM_VERT)      mode |=  ECL_VERT;
  if (jmode & fim_JNIFIM_ORIGSUPP)  mode |=  ECL_ORIGSUPP;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  if (!(target & ISR_RULES)) appear = NULL;
  tabag = tbg_fromJObj(env, tracts, wgts, appear);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  eclat = eclat_create(target, supp, 100.0, conf,
                       (ITEM)zmin, (ITEM)zmax,
                       eval, agg, thresh, algo, mode);
  if (!eclat) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = eclat_data(eclat, tabag, 0, +2);
  if (r) eclat_delete(eclat, 1);    /* prepare data for eclat */
  if (r == -1) return ERR_MEM(env); /* and check for error */
  if (r <   0) return ERR_RTN(env); /* and for no items */

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create item set reporter */
  if (!isrep) { eclat_delete(eclat, 1); return ERR_MEM(env); }
  r = eclat_report(eclat, isrep);      /* prepare item set reporter */
  if (r)      { eclat_delete(eclat, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    eclat_delete(eclat, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (prune < ITEM_MIN) prune = ITEM_MIN;
  if (prune > ITEM_MAX) prune = ITEM_MAX;
  r = eclat_mine(eclat, (ITEM)prune, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  eclat_delete(eclat, 1);       /* delete the eclat miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_eclat() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_fpgrowth
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jdouble conf, jint zmin, jint zmax,
   jstring report, jint eval, jint agg, jdouble thresh, jint prune,
   jint algo, jint jmode, jintArray border, jobjectArray appear)
{                               /* --- FP-growth algorithm */
  int      mode = FPG_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  FPGROWTH *fpgrowth;           /* fpgrowth miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, SETSvRULES); /* translate target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if ((conf  < 0) || (conf > 100)) /* check rule confidence */
    return ERR_ARG(env, "invalid confidence");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;   /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (zmin   > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax   > ITEM_MAX) zmax = ITEM_MAX;
  eval = get_evalx(eval);       /* get and check evaluation measure */
  if (eval  < 0)    return ERR_ARG(env, "invalid evaluation measure");
  if (eval  <= RE_NONE) prune = ITEM_MIN;
  if (mode & fim_JNIFIM_INVBXS) eval |= FPG_INVBXS;
  agg  = get_agg(agg);          /* get and check aggregation mode */
  if (agg   < 0)    return ERR_ARG(env, "invalid aggregation mode");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = FPG_SIMPLE;  break;
    case fim_JNIFIM_FPG_SIMPLE:     algo = FPG_SIMPLE;  break;
    case fim_JNIFIM_FPG_COMPLEX:    algo = FPG_COMPLEX; break;
    case fim_JNIFIM_FPG_SINGLE:     algo = FPG_SINGLE;  break;
    case fim_JNIFIM_FPG_TOPDOWN:    algo = FPG_TOPDOWN; break;
    default: return ERR_ARG(env, "invalid FP-growth algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOFIM16)   mode &= ~FPG_FIM16;
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~FPG_PERFECT;
  if (jmode & fim_JNIFIM_NOSORT)    mode &= ~FPG_REORDER;
  if (jmode & fim_JNIFIM_NOHUT)     mode &= ~FPG_TAIL;
  if (jmode & fim_JNIFIM_ORIGSUPP)  mode |=  FPG_ORIGSUPP;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  if (!(target & ISR_RULES)) appear = NULL;
  tabag = tbg_fromJObj(env, tracts, wgts, appear);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  fpgrowth = fpg_create(target, supp, 100.0, conf,
                        (ITEM)zmin, (ITEM)zmax,
                        eval, agg, thresh, algo, mode);
  if (!fpgrowth) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = fpg_data(fpgrowth, tabag, 0, +2);
  if (r) fpg_delete(fpgrowth,1);/* prepare data for fpgrowth */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  r = fpg_report(fpgrowth, isrep);
  if (r) {                      /* prepare the item set reporter */
    fpg_delete(fpgrowth, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    isr_delete(isrep, 0); tbg_delete(tabag, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (prune < ITEM_MIN) prune = ITEM_MIN;
  if (prune > ITEM_MAX) prune = ITEM_MAX;
  r = fpg_mine(fpgrowth, (ITEM)prune, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  fpg_delete(fpgrowth, 1);      /* delete the fpgrowth miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_fpgrowth() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_sam
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint algo, jint jmode, jintArray border)
{                               /* --- SaM algorithm */
  int      mode = SAM_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  SAM      *sam;                /* split and merge miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, FCMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;   /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_eval(eval);        /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = SAM_BSEARCH; break;
    case fim_JNIFIM_SAM_BASIC:      algo = SAM_BASIC;   break;
    case fim_JNIFIM_SAM_BSEARCH:    algo = SAM_BSEARCH; break;
    case fim_JNIFIM_SAM_DOUBLE:     algo = SAM_DOUBLE;  break;
    case fim_JNIFIM_SAM_TREE:       algo = SAM_TREE;    break;
    default: return ERR_ARG(env, "invalid SaM algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOFIM16)   mode &= ~SAM_FIM16;
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~SAM_PERFECT;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  sam = sam_create(target, supp, 100.0, (ITEM)zmin, (ITEM)zmax,
                   0, -1.0, eval, thresh, algo, mode);
  if (!sam) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = sam_data(sam, tabag, +2); /* create a split and merge miner */
  if (r) sam_delete(sam, 1);    /* prepare data for split and merge */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    sam_delete(sam, 1); return ERR_MEM(env); }
   r = sam_report(sam, isrep);
  if (r) {                      /* prepare the item set reporter */
    sam_delete(sam, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    sam_delete(sam, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = sam_mine(sam, 8192);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  sam_delete(sam, 1);           /* delete the split and merge miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_sam() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_relim
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint algo, jint jmode, jintArray border)
{                               /* --- RElim algorithm */
  int      mode = REL_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  RELIM    *relim;              /* relim miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, FCMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;      /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_eval(eval);        /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = REL_BASIC; break;
    case fim_JNIFIM_RELIM_BASIC:    algo = REL_BASIC; break;
    default: return ERR_ARG(env, "invalid RElim algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOFIM16)   mode &= ~REL_FIM16;
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~REL_PERFECT;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  relim = relim_create(target, supp, 100.0, (ITEM)zmin, (ITEM)zmax,
                       0, -1.0, eval, thresh, algo, mode);
  if (!relim) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = relim_data(relim, tabag, +2);
  if (r) relim_delete(relim,1); /* prepare data for relim */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create item set reporter */
  if (!isrep) {                 /* create an item set reporter */
    relim_delete(relim, 1); return ERR_MEM(env); }
  r = relim_report(relim, isrep);
  if (r) {                      /* prepare the item set reporter */
    relim_delete(relim, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    relim_delete(relim, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = relim_mine(relim, 32);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  relim_delete(relim, 1);       /* delete the relim miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_relim() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_jim
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint covsim, jdoubleArray simps,
   jdouble sim, jint algo, jint jmode, jintArray border)
{                               /* --- JIM algorithm */
  int      mode = JIM_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  JIM      *jim;                /* jim miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */
  double   csps[8] =            /* cover similarity parameters */
           { 1, 0, 0, 0,  0, -1, 1, 0 };
  double   *p;                  /* to access similarity parameters */
  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, FCMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;      /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (eval == fim_JNIFIM_SIZESIM) eval = JIM_SIZESIM;
  else eval = get_eval(eval);   /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  covsim = get_covsim(covsim);  /* translate similarity measure code */
  if (covsim < 0) return ERR_ARG(env, "invalid cover similarity");
  if (sim    < 0) return ERR_ARG(env, "invalid similarity threshold");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = JIM_BASIC; break;
    case fim_JNIFIM_JIM_BASIC:      algo = JIM_BASIC; break;
    default: return ERR_ARG(env, "invalid JIM algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~JIM_PERFECT;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  if ((covsim >= JIM_GENERIC) && simps) {
    memset(csps, 0, 8*sizeof(double));
    r = (*env)->GetArrayLength(env, simps);
    p = (*env)->GetDoubleArrayElements(env, simps, NULL);
    while (--r >= 0) csps[r] = p[r];
    (*env)->ReleaseDoubleArrayElements(env, simps, p, JNI_ABORT);
  }                             /* copy similarity parameters */
  jim = jim_create(target, supp, (ITEM)zmin, (ITEM)zmax,
                   eval, thresh, covsim, csps, sim, algo, mode);
  if (!jim) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = jim_data(jim, tabag, +2); /* create a Jaccard item set miner */
  if (r) jim_delete(jim,1);     /* and prepare data for jim */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create item set reporter */
  if (!isrep) {                 /* create an item set reporter */
    jim_delete(jim, 1); return ERR_MEM(env); }
  r = jim_report(jim, isrep);   /* prepare the item set reporter */
  if (r) { jim_delete(jim, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    jim_delete(jim, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = jim_mine(jim);            /* mine frequent item sets */
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  jim_delete(jim, 1);           /* delete the jim miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_jim() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_carpenter
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint algo, jint jmode, jintArray border)
{                               /* --- Carpenter algorithm */
  int      mode = CARP_DEFAULT; /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  CARP     *carp;               /* carpenter miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, CMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;     /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_eval(eval);        /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = CARP_AUTO;    break;
    case fim_JNIFIM_CARP_TABLE:     algo = CARP_TABLE;   break;
    case fim_JNIFIM_CARP_TIDLIST:   algo = CARP_TIDLIST; break;
    default: return ERR_ARG(env, "invalid Carpenter algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOPERFECT) mode &= ~CARP_PERFECT;
  if (jmode & fim_JNIFIM_REPOFILT)  mode |=  CARP_FILTER;
  if (jmode & fim_JNIFIM_ADDALL)    mode &= ~CARP_MAXONLY;
  if (jmode & fim_JNIFIM_NOCOLLATE) mode &= ~CARP_COLLATE;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  carp = carp_create(target, supp, 100.0, (ITEM)zmin, (ITEM)zmax,
                     eval, thresh, algo, mode);
  if (!carp) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = carp_data(carp, tabag, +2);
  if (r) carp_delete(carp, 1);  /* prepare data for carpenter */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    carp_delete(carp, 1); return ERR_MEM(env); }
  r = carp_report(carp, isrep);
  if (r) {                      /* prepare the item set reporter */
    carp_delete(carp, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    carp_delete(carp, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = carp_mine(carp);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  carp_delete(carp, 1);         /* delete the carpenter miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_carpenter() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_ista
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint eval, jdouble thresh, jint algo, jint jmode, jintArray border)
{                               /* --- IsTa algorithm */
  int      mode  = ISTA_DEFAULT;/* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  ISTA     *ista;               /* ista miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target, CMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;     /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  eval = get_eval(eval);        /* get and check evaluation measure */
  if (eval   < 0) return ERR_ARG(env, "invalid evaluation measure");
  switch (algo) {               /* evaluate the algorithm code */
    case fim_JNIFIM_AUTO:           algo = ISTA_AUTO;     break;
    case fim_JNIFIM_ISTA_PREFIX:    algo = ISTA_PREFIX;   break;
    case fim_JNIFIM_ISTA_PATRICIA:  algo = ISTA_PATRICIA; break;
    default: return ERR_ARG(env, "invalid IsTa algorithm variant");
  }                           /* get the internal algorithm code */
  if (jmode & fim_JNIFIM_NOPRUNE)   mode &= ~ISTA_PRUNE;
  if (jmode & fim_JNIFIM_REPOFILT)  mode |=  ISTA_FILTER;

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  ista = ista_create(target, supp, 100.0, (ITEM)zmin, (ITEM)zmax,
                     eval, thresh, algo, mode);
  if (!ista) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = ista_data(ista, tabag, +2);
  if (r) ista_delete(ista, 1);  /* prepare data for ista */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag));
  if (!isrep) {                 /* create an item set reporter */
    ista_delete(ista, 1); return ERR_MEM(env); }
  r = ista_report(ista, isrep);
  if (r) {                      /* prepare the item set reporter */
    ista_delete(ista, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, target) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    ista_delete(ista, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  r = ista_mine(ista);
  if (r >= 0) r = repterm(env, &data, isrep, report, target);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  ista_delete(ista, 1);         /* delete the ista miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_ista() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_apriacc
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jdouble supp, jint zmin, jint zmax, jstring report,
   jint stat, jdouble siglvl, jint prune, jint jmode, jintArray border)
{                               /* --- Apriori algorithm (Accretion) */
  int      mode = APR_DEFAULT;  /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  APRIORI  *apriori;            /* apriori miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  if (zmin   <  0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   <  0) zmax = ITEM_MAX;  /* check the size range */
  if (zmax   <  zmin) return ERR_ARG(env, "zmax must be >= zmin");
  stat = get_stat(stat);        /* get and check evaluation measure */
  if (stat   <  0) return ERR_ARG(env, "invalid statistic");
  if (jmode & fim_JNIFIM_INVBXS) stat |= APR_INVBXS;
  if (siglvl <= 0) return ERR_ARG(env, "siglvl must be > 0");

  /* --- get and prepare transactions --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get transactions, create miner */
  apriori = apriori_create(ISR_MAXIMAL, supp, 100.0, 100.0,
                           (ITEM)zmin, (ITEM)zmax,
                           stat, APR_MAX, siglvl, APR_AUTO, mode);
  if (!apriori) { tbg_delete(tabag, 1); return ERR_MEM(env); }
  r = apriori_data(apriori, tabag, 0, +2);
  if (r) apriori_delete(apriori, 1);   /* prepare data for apriori */
  if (r == -1) return ERR_MEM(env);    /* and check for error */
  if (r <   0) return ERR_RTN(env);    /* and for no items */

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create an item set reporter */
  if (!isrep) { apriori_delete(apriori, 1); return ERR_MEM(env); }
  r = apriori_report(apriori, isrep);  /* prepare item set reporter */
  if (r)      { apriori_delete(apriori, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, ISR_MAXIMAL) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    apriori_delete(apriori, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (prune < ITEM_MIN) prune = ITEM_MIN;
  if (prune > ITEM_MAX) prune = ITEM_MAX;
  r = apriori_mine(apriori, (ITEM)prune, 0.01, 0);
  if (r >= 0) r = repterm(env, &data, isrep, report, ISR_MAXIMAL);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  apriori_delete(apriori, 1);   /* delete the apriori miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_apriacc() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_accretion
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jdouble supp, jint zmin, jint zmax, jstring report,
   jint stat, jdouble siglvl, jint maxext, jint jmode, jintArray border)
{                               /* --- Accretion algorithm */
  int      mode  = ACC_DEFAULT; /* operation mode/flags */
  TABAG    *tabag;              /* transaction bag */
  ISREPORT *isrep;              /* item set reporter */
  ACCRET   *accret;             /* accretion miner */
  REPDATA  data;                /* data for item set reporting */
  int      r;                   /* result of function call */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  if (zmin < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax < 0) zmax = ITEM_MAX;   /* check the size range */
  if (zmax < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  stat = get_stat(stat);        /* get and check evaluation measure */
  if (stat < 0) return ERR_ARG(env, "invalid statistic");
  if (mode & fim_JNIFIM_INVBXS) stat |= ACC_INVBXS;
  if (siglvl <= 0) return ERR_ARG(env, "siglvl must be > 0");
  if (maxext < 0)               /* a negative value means that */
    maxext = ITEM_MAX;          /* there is no limit on extensions */

  /* --- create transaction bag --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env);  /* get the given transactions */
  accret = accret_create(ISR_MAXIMAL, supp, 100.0,
                         (ITEM)zmin, (ITEM)zmax, stat, siglvl, mode);
  r = accret_data(accret, tabag, +2);
  if (r) accret_delete(accret, 1);  /* prepare data for accretion */
  if (r == -1) return ERR_MEM(env);
  if (r <   0) return ERR_RTN(env);

  /* --- create item set reporter --- */
  isrep = isr_create(tbg_base(tabag)); /* create item set reporter */
  if (!isrep) { accret_delete(accret, 1); return ERR_MEM(env); }
  r = accret_report(accret, isrep);    /* prepare item set reporter */
  if (r)      { accret_delete(accret, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border) != 0)
  ||  (repinit(env, &data, isrep, report, ISR_MAXIMAL) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    accret_delete(accret, 1); return ERR_MEM(env); }

  /* --- run frequent item set mining --- */
  if (maxext > ITEM_MAX) maxext = ITEM_MAX;
  r = accret_mine(accret, (ITEM)maxext);
  if (r >= 0) r = repterm(env, &data, isrep, report, ISR_MAXIMAL);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  accret_delete(accret, 1);     /* delete the accretion miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created result */
}  /* Java_fim_JNIFIM_accretion() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_genpsp
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint cnt, jint surr, jint seed, jint cpus, jintArray ctrl)
{                               /* --- generate a pattern spectrum */
  PATSPEC      *psp = NULL;     /* created pattern spectrum */
  jobjectArray jpsp = NULL;     /* created Java pattern spectrum */
  SUPP         smin = 2;        /* minimum support as an integer */
  TABAG        *tabag;          /* created transaction bag */
  double       wgt;             /* total transaction weight */
  const jchar  *fmt;            /* report format/pat. spectrum flag */
  JDATA        jd;              /* Java environment data */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate the function arguments --- */
  target = get_target(target, FCMSETS); /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin   < 0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax   < 0) zmax = ITEM_MAX;      /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (cnt   <= 0) cnt = 1;      /* check the number of data sets */
  surr   = get_surr(surr);      /* translate the surrogate code */
  if (surr  <  0) return ERR_ARG(env, "invalid surrogate method");
  if (surr  <= 0) cnt = 1;      /* adapt the number of data sets */

  /* --- generate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env); /* get the given transactions */
  if ((surr == 3) && !tbg_istab(tabag)) {
    tbg_delete(tabag, 1);       /* if shuffle surrogates requested */
    return ERR_ARG(env, "for shuffle surrogates "
                        "transactions must form a table");
  }                             /* check for table derived data */
  wgt  = tbg_wgt(tabag);        /* get the total transaction weight */
  supp = (supp >= 0) ? supp/100.0 *(double)wgt *(1-DBL_EPSILON) : -supp;
  smin = (SUPP)ceilsupp(supp);  /* compute absolute support value */
  jd.env = env; jd.ctrl = ctrl; /* fill java reporting data structure */
  psp  = fpg_genpsp(tabag, target, (SUPP)smin, (ITEM)zmin, (ITEM)zmax,
                    FPG_SIMPLE, FPG_DEFAULT, (size_t)cnt, surr,
                    (long)seed, cpus, repfn, &jd);
  if (psp) {                    /* estimate a pattern spectrum */
    fmt = (*env)->GetStringChars(env, report, NULL);
    jpsp = psp_toJObj(env, psp, 1.0/(double)cnt, fmt[0]);
    (*env)->ReleaseStringChars(env, report, fmt);
    psp_delete(psp);            /* turn it into a Java object and */
  }                             /* delete the pattern spectrum */
  tbg_delete(tabag, 1);         /* delete the created train set */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (!jpsp) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return jpsp;                  /* return the created Java object */
}  /* Java_fim_JNIFIM_genpsp() */

/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_fim_JNIFIM_estpsp
  (JNIEnv *env, jclass objc, jobjectArray tracts, jintArray wgts,
   jint target, jdouble supp, jint zmin, jint zmax, jstring report,
   jint equiv, jdouble alpha, jint smpls, jint seed)
{                               /* --- estimate a pattern spectrum */
  PATSPEC      *psp = NULL;     /* created pattern spectrum */
  jobjectArray jpsp = NULL;     /* created Java pattern spectrum */
  SUPP         smin = 2;        /* minimum support as an integer */
  TABAG        *tabag;          /* created transaction bag */
  double       wgt;             /* total transaction weight */
  const jchar  *fmt;            /* to evaluate the report string */

  assert(env);                  /* check the function arguments */
  if (!tracts) return ERR_NULL(env);

  /* --- evaluate the function arguments --- */
  target = get_target(target, ISR_SETS); /* translate target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (zmin  <  0) return ERR_ARG(env, "zmin must be >= 0");
  if (zmax  <  0) zmax = ITEM_MAX;       /* check the size range */
  if (zmax  <  zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (equiv <= 0) equiv = 1; /* check the number of surrogates */
  if (smpls <= 0) return ERR_ARG(env, "smpls must be > 0");
  if (seed  == 0) seed = (jint)time(NULL);

  /* --- estimate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  tabag = tbg_fromJObj(env, tracts, wgts, NULL);
  if (!tabag) return ERR_RTN(env);    /* get the given transactions */
  wgt  = tbg_wgt(tabag);        /* and the total transaction weight */
  supp = (supp >= 0) ? supp/100.0 *(double)wgt *(1-DBL_EPSILON) : -supp;
  smin = (SUPP)ceilsupp(supp);  /* compute absolute support value */
  if (tbg_recode(tabag, smin, -1, -1, -2) < 0) {
    tbg_delete(tabag, 1); return ERR_MEM(env); }
  tbg_filter(tabag, (ITEM)zmin, NULL, 0);
  psp = fpg_estpsp(tabag, target, (SUPP)supp, (ITEM)zmin, (ITEM)zmax,
                   (size_t)equiv, alpha, (size_t)smpls, seed);
  if (psp) {                    /* estimate a pattern spectrum */
    fmt = (*env)->GetStringChars(env, report, NULL);
    jpsp = psp_toJObj(env, psp, 1.0/(double)equiv, fmt[0]);
    (*env)->ReleaseStringChars(env, report, fmt);
    psp_delete(psp);            /* turn it into a Java object and */
  }                             /* delete the pattern spectrum */
  tbg_delete(tabag, 1);         /* delete the transaction bag */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (!jpsp) return ERR_MEM(env);
  sig_remove();                 /* remove the signal handler */
  return jpsp;                  /* return the created Java object */
}  /* Java_fim_JNIFIM_estpsp() */

/*--------------------------------------------------------------------*/
/* package : coconad                                                  */
/* class   : JNICoCo                                                  */
/* function: abort (static)                                           */
/*--------------------------------------------------------------------*/

JNIEXPORT void JNICALL Java_coconad_JNICoCo_abort
(JNIEnv *env, jclass objc, jint state)
{ sig_abort((int)state); }      /* --- set the abort state */
